# We cannot import Brick to top level because of bug --> Not possible to run outside agxViewer (debug) - fix on the way.
# from Brick.Physics import Component
import math
import agx
from typing import Tuple
import sys


class BrickUtils:

    @staticmethod
    def import_Brick():
        """
            Use:
            Brick = BrickUtils.import_Brick()
            Why:
            Because Brick can not be imported, and #import Brick actually takes quite some time (yet unknown - could be it's not cached)
            This convenience method speeds it up by forcing cache use.
        """
        assert "Brick" in sys.modules, "Brick was not available for import!"
        return sys.modules['Brick']

    @staticmethod
    def get_component_attribute(brick_component, attribute_name: str, default=None):
        """
            component - A Brick.Physics.Component to find attribute in
            attribute_name - the name of the attribute to find
            default - the value that should be returned if attribute_name does not exists in brick_component
        """
        # NOTE: There is unfortunately no way to ask a Brick.Physics.Component if it has an attribute, yet.
        try:
            return brick_component[attribute_name]
        except Exception:
            return default

    @staticmethod
    def roll_pitch_yaw(quat) -> Tuple[float, float, float]:
        """
            quat can be agx.Quat or Brick.Math.Quat
            returned tuple has order roll, pitch, yaw in radians
        """
#        from Brick.Math import Quat as BrickQuat
        # from
        # https://automaticaddison.com/how-to-convert-a-quaternion-into-euler-angles-in-python/
        if quat.__class__ == agx.Quat:
            x = quat.x()
            y = quat.y()
            z = quat.z()
            w = quat.w()
        else:  # Assume Brick.Math.Quat
            x = quat.X
            y = quat.Y
            z = quat.Z
            w = quat.W

        t0 = 2.0 * (w * x + y * z)
        t1 = 1.0 - 2.0 * (x * x + y * y)
        roll_rad = math.atan2(t0, t1)

        t2 = 2.0 * (w * y - z * x)
        t2 = 1.0 if t2 > +1.0 else t2
        t2 = -1.0 if t2 < -1.0 else t2
        pitch_rad = math.asin(t2)

        t3 = +2.0 * (w * z + x * y)
        t4 = +1.0 - 2.0 * (y * y + z * z)
        yaw_rad = math.atan2(t3, t4)

        return (roll_rad, pitch_rad, yaw_rad)



from typing import List
from agxclick import ClickObject, ClickRobot, BrickUtils


def has_click_configuration(scene_positioninput) -> bool:
    return BrickUtils.get_component_attribute(scene_positioninput, "clickobjects") is not None


def get_click_configuration(click_scene) -> List[ClickRobot]:
    """
    click_scene must be a Brick ClickScene root
    Will only return objects defined in clickobjects entry of ClickScene
    """
    objects = []
    import Brick.Robotics
    for object in click_scene["clickobjects"]:
        if isinstance(object, Brick.Robotics.Robot):
            objects.append(ClickRobot(object))
        else:
            objects.append(ClickObject(object))
    return objects


def find_robots_in_scene(scene_positioninput) -> List[ClickRobot]:
    """
    Looks for Brick.Robotics.Robot:s in scene
    Will not look for ClickScene entry
    """
    from Brick.Physics import Component

    def is_robot(robot: Component):
        # There can be robots in urdf that is not actual robots, which is why we check #arms
        import Brick.Robotics
        return isinstance(robot, Brick.Robotics.Robot) and len(robot.Arms) > 0

    def find_robots(component: Component):
        if is_robot(component):
            return [component]
        else:
            brick_robots = list()
            for comp in component.InternalComponents:
                brick_robots.extend(find_robots(comp))
            return brick_robots

    brick_robots = find_robots(scene_positioninput)
    robots = []
    for brick_robot in brick_robots:
        robot = ClickRobot(brick_robot)
        robots.append(robot)

    return robots



import time


class WallClock:

    def __init__(self):
        self.time = self.now()

    def now(self):
        return time.time()

    def time_since_init(self):
        return self.now() - self.time

    def timestamp(self):
        return self.now()

    def time_since(self, timestamp: float) -> float:
        return self.now() - timestamp



import agx
import agxOSG
import agxIO
import agxSDK
from typing import List
from agxBrickHosting import registerAgxSimulation
from agxPythonModules.utils.environment import create_or_set_script_context


class AgxApplication:

    def __init__(self, args: List[str]):
        self.init = agx.AutoInit()
        self.app = agxOSG.ExampleApplication()
        self.app.init(agxIO.ArgumentParser(args))
        # After app.init() Brick can be imported
        from Brick.AGXBrick import BrickSimulation
        self.sim = agxSDK.Simulation()
        self.app.initSimulation(self.sim, True)
        registerAgxSimulation(self.sim, "default")
        assert BrickSimulation.GetSimulation("default"), "Could not find brick simulation Default"
        assert BrickSimulation.Default, "Could not find brick simulation Default via c# attribute"
        # create_or_set_script_context line makes environment.application() and simulation() work
        create_or_set_script_context(self.sim, self.app, None)
        self.m_timer = agx.HighAccuracyTimer(True)
        self.have_graphics = None

    def stepApplication(self):
        """
        Steps application
        - Calls Listeners like KeyboardListener and FrameListeners
        - Renders graphics
        - Send/Receives RCS
        """
        # Cache the fact that we have graphics or not. Depends on stepApplication not being called until after setting up agxOSG
        if self.have_graphics is None:
            self.have_graphics = self.app.getViewer()
        if self.have_graphics:
            self.app.executeOneStepWithGraphics(self.m_timer)
        else:
            self.app.executeOneStepWithoutGraphics()

    def stepSimulation(self):
        """
        Steps simulation one timestep
        If self.app.getRealTimeSync() then sleep rest of timeslot
        """
        self.app.step()

    def stop(self):
        self.app.stop()
        agx.Runtime.instance().clear()

    def remote_viewer_exists(self) -> bool:
        return hasattr(self.app, "getController") and self.app.getController()

    def reset_scene(self, scene_to_reset):
        from Brick.Physics import ComponentLoader
        from Brick.AGXBrick import BrickSimulation

        ComponentLoader.RepositionComponent(scene_to_reset)
        BrickSimulation.Default.ResetAgx()
        if self.remote_viewer_exists():
            self.app.pushFrameToRemoteViewer()
        BrickSimulation.Default.SyncOutputParameters()



import signal
import sys
from agxclick import KeyboardListener, AgxApplication, ClickFrameListener, ApplicationStepListener, ResetBatchListener
from pclick import Server
from pclick.server import SizeCollector, SizeCollectorChanges
from typing import Any, Callable, List
import logging
from agxclick.wallclock import WallClock
from agxclick.graphics_throttler import GraphicsThrottler


class ClickApplication(AgxApplication):

    def signal_handler(self, sig, frame):
        self._logger.info("Stopping from SIGINT")
        self.on_stop()

    def __init__(self, args: List[str]):
        super().__init__(args)
        self._logger = logging.getLogger(__file__)
        self._stop_application = False
        self.args = self.parse_arguments(args)
        self.simulation_stepping_enabled = self.args.start_paused is False
        self.disable_clicksync = self.args.disable_clicksync is True
        self._click_frame_listener = None
        self._reset_batch_listener = None
        self.application_step_listeners: List[ApplicationStepListener] = []
        self.server = None
        signal.signal(signal.SIGINT, self.signal_handler)

    def run(self, buildScene: Callable[[], Any]):
        """
        buildScene must return the top Brick.Physics.Component of the scene
        """
        self.add_listeners()
        scene = buildScene()
        self.update_listeners(scene)

        # Store setting so we can restore it if overridden
        self._realtime_sync = self.app.getRealTimeSync()
        # We must control stepping ourselves to be able to drive steps from controller, and sync realtime if wanted
        self.app.setAutoStepping(False)
        # Since we drive the graphics from here, disable ExampleApplication graphics throttling
        self.app.setTargetFPS(0)

        self._logger.info(f"Entering main loop {'with' if self.app.getViewer() else 'without'} graphics targeting {self.args.framerate} Hz")
        self._logger.info(f"Realtime sync is {self.app.getRealTimeSync()}")
        if self.disable_clicksync:
            self._logger.info("Simulation will not wait for controlmessages")
        else:
            self._logger.info("Pausing simulation until handshake is completed, then each controlmessage will step simulation once")

        if self.args.profile:
            import cProfile
            profile = cProfile.Profile()
            profile.enable()

        num_frames, wall_clock = self.mainloop()
        # Release license seat - important if docker/floating license.
        self.stop()

        if self.args.profile:
            profile.disable()
            if self.args.profileFile == "":
                from pstats import SortKey, Stats
                ps = Stats(profile).sort_stats(SortKey.CUMULATIVE)
                ps.print_stats()
            else:
                profile.dump_stats(self.args.profileFile)

        self.report_timing(num_frames, wall_clock)

    def mainloop(self):
        num_frames = 0
        wall_clock = WallClock()
        gfx_throttler = GraphicsThrottler(self.args.framerate)
        click_sync_enabled = not self.disable_clicksync
        while not self._stop_application:
            # Reset wall_clock when handshake completed if syncing simulation step with controlmessage
            if click_sync_enabled and not self._click_frame_listener.handshake_completed():
                wall_clock = WallClock()
                num_frames = 0

            for listener in self.application_step_listeners:
                listener.preFrame(self.sim.getClock().getTime())
            click_ready_for_simstep = self.disable_clicksync or self._click_frame_listener.can_step_simulation()
            if click_ready_for_simstep and self.simulation_stepping_enabled:
                self.stepSimulation()
            if (gfx_throttler.render_graphics(wall_clock)):
                self.stepApplication()
                self.enforce_step_realtime_settings()
                num_frames += 1
            if self.args.stopAfterFrame and self.sim.getClock().getFrame() >= self.args.stopAfterFrame:
                self._stop_application = True
            if self.args.stopAfter and self.sim.getClock().getTime() >= self.args.stopAfter:
                self._stop_application = True
            if self.args.trace_sizes and self.server.size_collector.is_updated:
                print(f"Received {self.server.size_collector.recv_size} from client, Sent {self.server.size_collector.send_size} bytes to client")

        return num_frames, wall_clock

    def report_timing(self, num_frames, wall_clock):
        simsteps = self.sim.getClock().getFrame()
        walltime = wall_clock.time_since_init()
        self._logger.info(f"Rendered {num_frames} frames and {simsteps} simulation steps, received {self._click_frame_listener.num_controls_received} control messages")
        if (simsteps != self._click_frame_listener.num_controls_received):
            missed_simsteps = simsteps - self._click_frame_listener.num_controls_received
            self._logger.warn(f"Controller could not keep up with simulation! {missed_simsteps} simulation steps taken without control message")
        self._logger.info(f"simulated time: {self.sim.getClock().getTime()} Wall clock time: {walltime}")
        self._logger.info(f"Wallclock sim freq: {simsteps/walltime:.1f} Hz Wallclock framerate: {num_frames/walltime:.1f}")
        return simsteps, walltime

    def parse_arguments(self, args):
        from argparse import ArgumentParser
        parser = ArgumentParser(args)
        parser.add_argument('--addr', metavar='<addr>', type=str, default="tcp://*:5555",
                            help=f'set server address. Ie ipc:///tmp/click.ipc, default is tcp://*:5555')
        parser.add_argument('--batch', type=float, default=None,
                            help="Enable automatic restart of the scene after the specified number of seconds, with updated values for Brick BatchVariables and ParameterSpace variables")
        parser.add_argument('--disableClickSync', dest='disable_clicksync', action="store_true",
                            help="Do not sync each simulation step with click client - simulation will run without waiting for control messages")
        parser.add_argument('--framerate', type=int, default=60,
                            help="Specify target framerate in fps. 0 means match simulation framerate. default is 60. Recommended is 30-60. Only affects agxViewer, typically no speedup in browser")
        parser.add_argument('--profile', dest='profile', action="store_true", help="CProfile main loop and print results")
        parser.add_argument('--profileFile', type=str, default="", help="Write profile data to binary file (for snakeviz) instead of stdout")
        parser.add_argument('--startPaused', dest='start_paused', action="store_true", help="Start with simulation paused")
        parser.add_argument('--stopAfter', type=float, default=None, help="Stop when this simulation time is reached")
        parser.add_argument('--stopAfterFrame', type=int, default=None, help="Stop when this number of simulation steps has been executed")
        parser.add_argument('--trace-sizes', action='store_true',
                            help=f'print size of what is sent/received')
        args, _ = parser.parse_known_args(args)
        return args

    def on_exception(self, ex: Exception):
        self._stop_application = True
        self._logger.warning(f"on_exception called with {ex}")

    def on_stop(self):
        self._stop_application = True
        self.stop()
        self._click_frame_listener.stop()

    def on_keyboard_reset(self):
        self._logger.info("Resetting scene")
        self.reset_scene(self._scene)
        self._click_frame_listener.send_reset()

    def reset_scene(self, scene_to_reset):
        if (self._reset_batch_listener is not None):
            self._reset_batch_listener.prepare_for_next_batch_state()

        super().reset_scene(scene_to_reset)

    def on_reset_message(self):
        assert self._scene
        self.reset_scene(self._scene)

    def on_toggle_simulation_stepping(self):
        self.simulation_stepping_enabled = not self.simulation_stepping_enabled
        state_text = "running" if self.simulation_stepping_enabled else "paused"
        self._logger.info(f"Simulation is {state_text}")

    def update_listeners(self, scene):
        """
        We need to add listeners before possible listeners added by buildScene, which is why these updates are needed
        """
        self._scene = scene
        for listener in self.application_step_listeners:
            listener.update_scene(scene)

    def add_listeners(self):
        self._logger.info(f"Starting server on {self.args.addr}")
        size_collector = SizeCollectorChanges() if self.args.trace_sizes else SizeCollector()
        self.server = Server(self.args.addr, size_collector)
        self._click_frame_listener = ClickFrameListener(scene=None, app=self.app,
                                                        on_stop=self.on_stop,
                                                        on_exception=self.on_exception,
                                                        on_reset=self.on_reset_message,
                                                        server=self.server
                                                        )
        self.application_step_listeners.append(self._click_frame_listener)
        if self.args.batch is not None:
            self._reset_batch_listener = ResetBatchListener(scene=None, batch_time=self.args.batch,
                                                            on_batch_end=self.on_keyboard_reset)
            self.application_step_listeners.append(self._reset_batch_listener)
        keyboardListener = KeyboardListener(on_stop=self.on_stop,
                                            on_reset=self.on_keyboard_reset,
                                            on_toggle_stepping=self.on_toggle_simulation_stepping
                                            )
        self.sim.add(keyboardListener)

    def enforce_step_realtime_settings(self):
        if self.app.getAutoStepping():
            self._logger.info(f"Something (SimulationView?) enabled autostepping, disabling it again...")
            self.app.setAutoStepping(False)
        if self.app.getRealTimeSync() is not self._realtime_sync:
            self._logger.info(f"Something (SimulationView?) changed realtime sync, forcing it to {self._realtime_sync} again...")
            self.app.setRealTimeSync(self._realtime_sync)



from typing import Any, List, Tuple, Dict
from agxclick import BrickUtils
import logging
import itertools
from pclick import ValueType


class ClickObject():
    def __init__(self, component):
        """
        component is a Brick.Component
        """
        self.component = component
        self.name = component.Name
        self.logger = logging.getLogger(__file__)

    def position(self) -> Tuple[float, float, float]:
        pos = self.component.Transform.Position  # Brick.Math.Vec3
        return (pos.X, pos.Y, pos.Z)

    def rpy(self) -> Tuple[float, float, float]:
        return BrickUtils.roll_pitch_yaw(self.component.Transform.Rotation)

    def is_robot(self):
        return False

    def __str__(self):
        def classname(cls):
            return f"{cls.__class__.__module__}.{cls.__class__.__name__}"

        def values(signals):
            return list(map(lambda signal: signal.GetData(), signals))

        def signalstr(signals):
            return f"{classname(signals[0])}: {values(signals)}"

        def eventstr(tuple: Tuple[str, Any]):
            key, value = tuple
            return f"{key}: {value.GetData()}"

        return f"name: {self.name} with component {self.component.__class__}"


class ClickRobot(ClickObject):
    def __init__(self, brick_robot):
        """
        brick_robot must be instance of Brick.Robotics.Robot, typically a Brick.AGXBrick.UrdfModel
        Throws exception if
          - Any unknown signal (unknown=not implemented interpretation of) is found in robot
          - Number of input signals or sensors does not match number of joints
        """
        super().__init__(brick_robot)
        Brick = BrickUtils.import_Brick()
        assert isinstance(brick_robot, Brick.Robotics.Robot)
        self.brickrobot = brick_robot
        self.input_signals: List[Brick.Signal.LockPositionInput | Brick.Signal.ForceInput | Brick.Signal.VelocityInput] = []
        self.torque_sensors: List[Brick.Signal.ForceScalarOutput] = []    # Nm
        self.angle_sensors: List[Brick.Signal.MotorAngleOutput] = []     # DEG in Brick
        self.velocity_sensors: List[Brick.Signal.MotorVelocityOutput] = []  # RAD/s in Brick
        self.num_joints = len(self.joints())
        self.control_event_dict = {}
        self.sensors: Dict[str, List[Brick.Signal.Output]] = {}
        self._populate_signals()
        self.validate()

    def is_robot(self):
        return True

    def validate(self):
        """
        Throws exception if Robot is not in a valid state
        """
        assert len(self.name) > 1, "Invalid robot name '{self.name}'"
        assert self.num_joints == len(self.input_signals), f"Number of input_signals {len(self.input_signals)} did not match number of joints {self.num_joints}"
        if (self.torque_sensors):
            assert self.num_joints == len(self.torque_sensors), f"Number of torque_sensors {len(self.torque_sensors)} did not match number of joints {self.num_joints}"
        if (self.angle_sensors):
            assert self.num_joints == len(self.angle_sensors), f"Number of angle_sensors {len(self.angle_sensors)} did not match number of joints {self.num_joints}"
        if (self.velocity_sensors):
            assert self.num_joints == len(self.velocity_sensors), f"Number of velocity_sensors {len(self.velocity_sensors)} did not match number of joints {self.num_joints}"

        assert None not in self.joint_protocolrefs(), f"Missing protocolReference in robot {self.name}, refs are: {self.joint_protocolrefs()}"

    def joints(self) -> List[Any]:
        def flatten(lists: List[List]) -> List:
            return list(itertools.chain.from_iterable(lists))
        return flatten(map(lambda arm: arm.Joints, self.brickrobot.Arms))

    def joint_protocolrefs(self) -> List[str]:
        """
        Returns protocolReference name of all joints in a sorted list
        """
        return list(map(lambda joint: joint.ProtocolReference, self.joints()))

    def control_event_names(self) -> List[str]:
        return self.control_event_dict.keys()

    def control_events(self) -> Dict:
        return self.control_event_dict

    def has_control_events(self):
        return bool(self.control_event_dict)

    def joint_sensors(self) -> List[Tuple[str, Any]]:
        """
            returns list with tuples (protocolReference, OutputSignal: Brick.Signal.*Output)
        """
        return list(map(lambda sensor: (sensor.Joint.ProtocolReference, sensor.GetOrCreateSignal()), self.brickrobot.Sensors))

    def controlType(self):
        """
        DEPRECATED: Use control_types()
        Returns how this robot expects to be controlled
        return a Brick.Signal.SignalBase, one of Brick.Signal.*Input, see https://brick:ien1ieh7Cithoohoh2AhNg0waingaigu@d2epgodm7v8ass.cloudfront.net/Introduction.html under Modules->Signal
            - Signal.LockPositionInput
            - Signal.MotorForceInput
            - Signal.MotorVelocityInput
        """
        control_types = set(self.control_types())
        assert len(control_types) > 0, f"Misconfiguration: Robot {self.name} has no input signals"
        assert len(control_types) == 1, f"Robot {self.name} has multiple controltypes - use control_types()"
        return control_types.pop()

    def control_types(self) -> List:
        """
        Returns how this robot expects to be controlled
        return a List[Brick.Signal.SignalBase], one of Brick.Signal.*Input, see https://brick:ien1ieh7Cithoohoh2AhNg0waingaigu@d2epgodm7v8ass.cloudfront.net/Introduction.html under Modules->Signal
            - Signal.LockPositionInput
            - Signal.MotorForceInput
            - Signal.MotorVelocityInput
        """
        return [s.__class__ for s in self.input_signals]

    def __str__(self):
        def classname(cls):
            return f"{cls.__class__.__module__}.{cls.__class__.__name__}"

        def values(signals):
            return list(map(lambda signal: signal.GetData(), signals))

        def types(signals):
            return list(map(lambda signal: classname(signal), signals))

        def signalstr(signals):
            if len(signals) == 0:
                return "[]"
            return f"{classname(signals[0])}: {values(signals)}"

        def eventstr(tuple: Tuple[str, Any]):
            key, value = tuple
            return f"{key}: {value.GetData()}"

        return f"""name: {self.name}
    num_joints: {self.num_joints}
    jointnames: {self.joint_protocolrefs()}
    control input type: {self.controlType() if len(set(self.control_types())) == 1 else "Multiple"}
    {len(self.input_signals)} input_types: {types(self.input_signals)}
    {len(self.input_signals)} input_values: {values(self.input_signals)}
    {len(self.torque_sensors)} torque_sensors: {signalstr(self.torque_sensors)}
    {len(self.angle_sensors)} angle_sensors: {signalstr(self.angle_sensors)}
    {len(self.velocity_sensors)} velocity_sensors: {signalstr(self.velocity_sensors)}
    control events: {list(map(eventstr, self.control_events().items()))}
"""

    def _populate_signals(self):
        for signal in self.brickrobot.GetSignals():
            self._populate_signal(signal)

    def _sensor_of_or(self, signal, default_reply):
        for sensor in self.brickrobot.Sensors:
            if signal == sensor.GetOrCreateSignal():
                return sensor
        return default_reply

    def _sensor_of(self, signal):
        for sensor in self.brickrobot.Sensors:
            if signal == sensor.GetOrCreateSignal():
                return sensor
        raise Exception("Signal {signal} was not part of any sensor in {self.name}")

    def _populate_signal(self, signal):
        Brick = BrickUtils.import_Brick()
        sensor = self._sensor_of_or(signal, object)
        if isinstance(sensor, Brick.Robotics.Sensor.JointSensor):
            if signal.__class__ is Brick.Signal.MotorAngleOutput:
                self.angle_sensors.append(signal)
            elif signal.__class__ is Brick.Signal.MotorVelocityOutput:
                self.velocity_sensors.append(signal)
            elif isinstance(signal, Brick.Signal.ForceScalarOutput):
                self.torque_sensors.append(signal)
        elif isinstance(signal, (Brick.Signal.LockPositionInput,
                                 Brick.Signal.ForceInput,
                                 Brick.Signal.VelocityInput)):
            self.input_signals.append(signal)
        elif signal.__class__ is Brick.Signal.AdhesiveForceInput:
            shortname = signal['name']
            if shortname == "":
                shortname = str(signal._ModelValuePath).rsplit('.', 1)[-1]
            self.logger.info(f"Configuring signal {signal._ModelValuePath} as controlEvent, will accept {shortname} in controlmessage for {self.name}")
            self.control_event_dict[shortname] = signal
        elif signal.__class__ is Brick.Signal.ComponentBoolInput or \
                signal.__class__ is Brick.Signal.ComponentBoolOutput:
            shortname = signal['name']
            if shortname == "":
                shortname = str(signal._ModelValuePath).rsplit('.', 1)[-1]
            self.logger.info(f"Configuring signal {signal._ModelValuePath} as controlEvent, will accept {shortname} in controlmessage for {self.name}")
            self.control_event_dict[shortname] = signal
        elif isinstance(signal, (Brick.Signal.ConnectorVectorOutput,
                                 Brick.Signal.PositionScalarOutput,
                                 Brick.Signal.VelocityScalarOutput,
                                 Brick.Signal.ForceScalarOutput)):
            id = self._sensor_of(signal)['protocolReference']
            assert id, f"Missing protocolReference in robot {self.name} sensor {self._sensor_of(signal)['name']}"
            self.logger.info(f"Configuring signal {signal._ModelValuePath} as sensor, with name {id} in {self.name}")
            if id in self.sensors:
                self.sensors[id].append(signal)
            else:
                self.sensors[id] = [signal]
        else:
            raise Exception(f"Unrecognized signal in {self.name}: {signal._ModelType}, class: {signal.__class__}")



import queue
import agxSDK
from typing import List, Callable
from agxclick import MessageFactory, ApplicationStepListener, update_robots_from_message, find_robots_in_scene, has_click_configuration, ClickObject, noop
from pclick import Server, HandshakeInitMessageType, ControlMessageType, ResetMessageType, SensorRequestMessageType, ErrorMessageType
from pclick import ControlMessage, MessageFactory as ProtoMessageFactory
import logging
from queue import SimpleQueue
from enum import Enum, auto

from agxclick import get_click_configuration


class States(Enum):
    INVALID = auto()

    RECV_HANDSHAKE = auto()
    RECV = auto()

    READ_CONTROLS = auto()
    SEND_SENSORS = auto()
    SEND_RESET = auto()

    def valid(self):
        return self is not self.INVALID

    def can_receive(self):
        return self is self.RECV or self is self.RECV_HANDSHAKE or self is self.SEND_RESET

    def can_send(self):
        return self is self.SEND_SENSORS or self is self.READ_CONTROLS


# class ClickFrameListener(agxOSG.ExampleApplicationListener):
class ClickFrameListener(ApplicationStepListener):
    def __init__(self, scene, app,
                 on_stop: Callable[[], None] = noop,
                 on_exception: Callable[[Exception], None] = noop,
                 on_reset: Callable[[], None] = noop,
                 server: Server = None):
        """
        scene is the top Brick.Component of the scene
        app is the running ExampleApplication
        sim is the agx simulation
        on_stop is an optional function to call on ErrorMessage, ie when stopping application
        on_exception is an optional function to call when exception occurs
        on_reset is an optional function to call when a ResetMessage is received
        server is an optional server to use instead of the default Server("tcp://*:5555")
        """
        super().__init__()
        self._logger = logging.getLogger(__file__)
        self._app = app
        self._on_stop = on_stop
        self._on_exception = on_exception
        self._on_reset = on_reset
        if scene is None:
            self._logger.warning("Scene is None, update_scene() must be called before first application step!")
            self._click_objects = None
        else:
            self.update_scene(scene)

        self._server = server if server else Server(f"tcp://*:5555")
        self.control_queue: SimpleQueue[ControlMessage] = SimpleQueue()
        self.click_event_listener = ClickEventListener(self._server, self._click_objects, self.control_queue, self._on_exception)
        self._app.getSimulation().addEventListener(self.click_event_listener, agxSDK.EventManager.HIGHEST_PRIORITY)
        self._app.getSimulation().addEventListener(ClickOutputEventListener(self.click_event_listener), agxSDK.EventManager.LOWEST_PRIORITY)
        self.num_controls_received = 0

    @property
    def _state(self):
        return self.click_event_listener.state

    @_state.setter
    def _state(self, new_state: States):
        self.click_event_listener.state = new_state

    def update_scene(self, scene_positioninput):
        if (has_click_configuration(scene_positioninput)):
            self._click_objects: List[ClickObject] = get_click_configuration(scene_positioninput)
        else:
            self._click_objects: List[ClickObject] = find_robots_in_scene(scene_positioninput)
        self.click_event_listener._click_objects = self._click_objects
        self._retrieve_initial_values()

    def _retrieve_initial_values(self):
        from Brick.AGXBrick import BrickSimulation
        BrickSimulation.Default.SyncOutputParameters()

    def can_step_simulation(self):
        return self._state is States.READ_CONTROLS

    def handshake_completed(self):
        return self._state is not States.RECV_HANDSHAKE

    def preFrame(self, time: float):
        try:
            if not self._state.can_receive():
                return
            request = self._server.recv()
            if request is None:
                return
            if request.messageType == HandshakeInitMessageType:
                self._logger.info("Received handshake init message")
                message = MessageFactory.handshake_message_from_objects(self._click_objects, self._app.getSimulation().getTimeStep())
                self._server.send(message)
                if self._state == States.SEND_RESET:
                    self._state = States.RECV
            elif self._state == States.SEND_RESET:
                self._send_resetmessage()
            elif request.messageType == SensorRequestMessageType:
                message = MessageFactory.sensor_message_from_objects(self._click_objects, self._app.getSimulation().getClock().getTime())
                self._logger.info(f"Received Sensor Request message, sending {message}")
                self._server.send(message)
            elif request.messageType == ControlMessageType:
                self.num_controls_received += 1
                self.control_queue.put(request)
                if self._state == States.RECV_HANDSHAKE:
                    self._logger.info("Handshake completed")
                self._state = States.READ_CONTROLS
            elif request.messageType == ErrorMessageType:
                self._logger.warning("Received Error message")
                self._on_stop()
            elif request.messageType == ResetMessageType:
                self._logger.info("Received Reset message")
                self._on_reset()
                self._server.send(ProtoMessageFactory.create_resetmessage())
            else:
                self._logger.warning(f"Ignoring message: \n{request}\nSending errormessage as reply")
                self._server.send(ProtoMessageFactory.create_errormessage())
        except Exception as ex:
            self._logger.info("Exception encountered - Stopping click messaging")
            self._state = States.INVALID
            self._on_exception(ex)
            raise ex

    def _send_resetmessage(self):
        self._logger.info("Sending reset message")
        self._server.send(ProtoMessageFactory.create_resetmessage())
        self._state = States.RECV

    def send_reset(self):
        if self._state.can_send():
            self._send_resetmessage()
        else:
            self._logger.info("Sending Reset as next Response (unless Handshake encountered)")
            self._state = States.SEND_RESET

        try:
            self.control_queue.get(block=False)
        except queue.Empty:
            self._logger.info("ControlMessage queue emptied because of Reset")
            assert self.control_queue.empty()

    def stop(self):
        self._server.stop()


class ClickEventListener(agxSDK.StepEventListener):
    """
    Responsible for retrieving messages and updating click objects.
    This should be done as early as possible in a simulation step, so set a high priority for this listener
    """
    def __init__(self, server: Server, click_objects: List[ClickObject], control_queue: SimpleQueue, on_exception: Callable[[Exception], None] = noop):
        """
        """
        super().__init__(agxSDK.StepEventListener.PRE_COLLIDE)
        self._server = server
        self._click_objects = click_objects
        self._control_queue: SimpleQueue = control_queue
        self._on_exception = on_exception
        self._logger = logging.getLogger(__file__)
        # The state is shared by both listeners
        self.state = States.RECV_HANDSHAKE

    def preCollide(self, time: float):
        if self.state is not States.READ_CONTROLS:
            return
        try:
            self._logger.debug(f"Updating click objects")
            request = self._control_queue.get(block=False)
            assert self._control_queue.empty()

            update_robots_from_message(self._click_objects, request)
            self.state = States.SEND_SENSORS
        except queue.Empty as ex:
            self._logger.fatal(f"No ControlMessage in queue")
            self.state = States.INVALID
            self._on_exception(ex)
            raise ex
        except Exception as ex:
            self._logger.fatal(f"Exception {str(ex)} encountered - Stopping click messaging")
            self.state = States.INVALID
            self._on_exception(ex)
            raise ex


class ClickOutputEventListener(agxSDK.StepEventListener):
    """
    Responsible for sending and creating messages from click objects.
    This should be done as late as possible in a simulation step to make sure all values are done propagating from AGX to Brick.
    Run with a low priority.

    The reason this is a separate class from ClickEventListener is because of the difference in call priority.
    This class is a proxy for a companioning ClickEventListener where all data is stored.
    """

    def __init__(self, input_listener: ClickEventListener):
        """
        input_listener is the companioning ClickEventListener where all data is stored.
        """
        super().__init__(agxSDK.StepEventListener.LAST_STEP)
        self._input_listener = input_listener

    def last(self, time: float):
        if self._input_listener.state is not States.SEND_SENSORS:
            return
        try:
            self._input_listener._logger.debug(f"Sending sensor message")
            response = MessageFactory.sensor_message_from_objects(self._input_listener._click_objects, time)
            self._input_listener._server.send(response)
            self._input_listener.state = States.RECV
        except Exception as ex:
            self._input_listener._logger.info(f"Exception encountered - Stopping click messaging")
            self._input_listener.state = States.INVALID
            self._input_listener._on_exception(ex)
            raise ex



import agxSDK
from typing import Callable
from agxPythonModules.utils.environment import application


class KEYS:
    n = 110
    e = 101
    r = 114


def noop(*args, **kwargs: None):
    pass


class KeyboardListener(agxSDK.GuiEventListener):

    def __init__(self,
                 on_stop: Callable[[], None] = noop,
                 on_reset: Callable[[], None] = noop,
                 on_toggle_stepping: Callable[[], None] = noop):
        """
        on_stop: callback for when Esc is pressed
        on_reset: callback for when n is pressed
        """
        super().__init__(agxSDK.GuiEventListener.KEYBOARD)
        self._on_stop = on_stop
        self._on_reset = on_reset
        self._on_toggle_stepping = on_toggle_stepping

    def keyboard(self, key, modKey, x, y, keyDown: bool):

        if not keyDown:
            return False

        if key == agxSDK.GuiEventListener.KEY_Escape:
            self._on_stop()
            return True

        if key == KEYS.n:
            self._on_reset()
            return True

        if key == KEYS.e:
            self._on_toggle_stepping()
            return True

        if key == KEYS.r:
            print("NOTE: Stepping simulation will fail if there is not a ControlMessage in queue from controller")
            return False

        return False



from agxclick.brick_utils import BrickUtils
from agxclick.click_robot import ClickRobot, ClickObject
from agxclick.brick_reader import find_robots_in_scene, get_click_configuration, has_click_configuration
from agxclick.message_factory import MessageFactory, update_robots_from_message
from agxclick.application_step_listener import ApplicationStepListener, noop
from agxclick.reset_batch_listener import ResetBatchListener
from agxclick.click_event_listener import ClickFrameListener
from agxclick.keyboard_listener import KeyboardListener
from agxclick.agx_application import AgxApplication
from agxclick.click_application import ClickApplication



class ApplicationStepListener():
    def __init__(self, scene=None) -> None:
        self._scene = scene

    def preFrame(self, time: float):
        pass

    def update_scene(self, scene):
        self._scene = scene


def noop(*args, **kwargs: None):
    pass



from agxclick.wallclock import WallClock

class GraphicsThrottler:

    def __init__(self, framerate: int):
        """
        If framerate is 0 or negative -> render_graphics always returns True
        """
        self.framerate = framerate
        self.last_render_timestamp = 0

    def render_graphics(self, wall_clock: WallClock, auto_reset=True) -> bool:
        """
        """
        if self.framerate <= 0:
            return True
        render = wall_clock.time_since(self.last_render_timestamp) >= 1.0 / self.framerate
        if render and auto_reset:
            self.rendered(wall_clock)
        return render

    def rendered(self, wall_clock: WallClock):
        self.last_render_timestamp = wall_clock.timestamp()



import math
from typing import List
from pclick import HandshakeMessage, SensorMessage, ValueType, MessageFactory as ProtoMessageFactory
from agxclick import ClickRobot, BrickUtils
from agxclick.click_robot import ClickObject


def _identity(x):
    return x


def _set_converted_signal_value(targets: List, source: List[float]):
    """
    Set signal values
    Converts Angle type values to degrees
    target is a list of Brick.Signal
    """
    for i, target in enumerate(targets):
        control_type = MessageFactory.to_click_control_type(target.__class__)
        if control_type == ValueType.Angle:
            conversion_function = math.degrees
        else:
            conversion_function = _identity

        target.SetData(conversion_function(source[i]))


def update_robots_from_message(robots: List[ClickRobot], controlmessage):
    """
    Updates a list of robots from a controlmessage
    """
    for robot in filter(lambda object: object.is_robot(), robots):
        # NOTE: We could output warning and return, but we prefer to fail to prevent user missing the problem
        assert robot.name in controlmessage.objects, f"Robot {robot.name} not found in controlmessage - client probably sent wrong name"
        control = controlmessage.objects[robot.name]
        if len(control.values):
            validate_message(controlmessage, robot, control.values)
            _set_converted_signal_value(robot.input_signals, control.values)
        else:
            # For backward compatibility and possibly performance we still support these:
            control_type = MessageFactory.to_click_control_type(robot.controlType())
            if control_type == ValueType.Angle:
                validate_message(controlmessage, robot, control.angles)
                _set_converted_signal_value(robot.input_signals, control.angles)
            elif control_type == ValueType.AngleVelocity:
                validate_message(controlmessage, robot, control.angleVelocities)
                _set_converted_signal_value(robot.input_signals, control.angleVelocities)
            elif control_type == ValueType.Torque:
                validate_message(controlmessage, robot, control.torques)
                _set_converted_signal_value(robot.input_signals, control.torques)
            else:
                raise Exception(f"Updating robot from controltype {robot.controlType()} has not been implemented")
        for key, enabled in control.controlEvents.items():
            if type(robot.control_events()[key].GetData()) == bool:
                robot.control_events()[key].SetData(enabled)
            else:
                robot.control_events()[key].SetData(100.0 if enabled else 0.0)


def validate_message(controlmessage, robot, values):
    assert len(robot.input_signals) == len(values), f"Mismatching number of values for {robot.name} in controlmessage, got {len(values)} should be {len(robot.input_signals)} - {controlmessage}"


class MessageFactory:

    @classmethod
    def to_click_control_types(cls, types: List) -> ValueType:
        return list(map(lambda x: cls.to_click_control_type(x.__class__), types))

    @classmethod
    def to_click_control_type(cls, type) -> ValueType:
        """
        Convert a Brick type to the corresponding click ValueType
        type is a Brick.Signal.*Input
        """
        if not hasattr(cls, "typemap"):
            Brick = BrickUtils.import_Brick()
            cls.typemap = {
                Brick.Signal.LockPositionInput: ValueType.Angle,
                Brick.Signal.VelocityInput: ValueType.AngleVelocity,
                Brick.Signal.ForceInput: ValueType.Torque,
                Brick.Signal.MotorVelocityInput: ValueType.AngleVelocity,
                Brick.Signal.FixedVelocityEngineInput: ValueType.AngleVelocity,
                Brick.Signal.EngineTorqueInput: ValueType.Torque,
                Brick.Signal.MotorForceInput: ValueType.Torque,
                # NOTE: AdhesiveForceInput is mapped to Activated to support suction cup, should be more generic, ie might not always want to map Adhesive to bool
                Brick.Signal.AdhesiveForceInput: ValueType.Activated,
                Brick.Signal.ForceVectorOutput: ValueType.Force,
                Brick.Signal.TorqueVectorOutput: ValueType.DirectionalTorque,
                Brick.Signal.RotatingBodyAngleOutput: ValueType.Angle,
                Brick.Signal.RotatingBodyVelocityOutput: ValueType.AngleVelocity,
                Brick.Signal.FixedVelocityEngineTorqueOutput: ValueType.Torque,
                Brick.Signal.ComponentBoolInput: ValueType.Activated,
                Brick.Signal.ComponentBoolOutput: ValueType.Activated,
            }
        return cls.typemap[type]

    @classmethod
    def _degs_to_radians(cls, signals) -> List[float]:
        """
        Convert a list of Brick signals in degrees to radians
        return a Brick.Signal.*Input type
        """
        return list(map(lambda signal: math.radians(signal.GetData()), signals))

    @classmethod
    def _signals_to_floats(cls, signals) -> List[float]:
        """
        Convert a list of Brick signals to List of floats
        return a Brick.Signal.*Input type
        """
        return list(map(lambda signal: cls._signal_to_floats(signal), signals))

    @classmethod
    def _signal_to_floats(cls, signal):
        """
        Convert data of Brick.Signal.ConnectorVectorOutput Brick.Math.Vec3 to list of 3 floats
        """
        Brick = BrickUtils.import_Brick()
        data = signal.GetData()
        if data.__class__ is Brick.Math.Vec3:
            return [data.X, data.Y, data.Z]
        return data

    @classmethod
    def handshake_message_from_objects(cls, objects: List[ClickObject], timeStep) -> HandshakeMessage:
        """
        Creates a HandshakeMessage from a list of robots
        """
        handshake = ProtoMessageFactory.create_handshake()
        handshake.controlType = cls.check_uniform_controltype(objects)
        handshake.simSettings.timeStep = timeStep
        for robot in objects:
            object = handshake.objects[robot.name]
            if robot.is_robot():
                object.controlsInOrder.extend(robot.joint_protocolrefs())
                object.controlTypesInOrder.extend(cls.to_click_control_types(robot.input_signals))
                object.jointSensorsInOrder.extend(robot.joint_protocolrefs())
                jointsensors = list()
                if len(robot.angle_sensors) > 0:
                    jointsensors.append(ValueType.Angle)
                if len(robot.velocity_sensors) > 0:
                    jointsensors.append(ValueType.AngleVelocity)
                if len(robot.torque_sensors) > 0:
                    jointsensors.append(ValueType.Torque)
                object.jointSensors.extend(jointsensors)
                for name, event in robot.control_events().items():
                    object.controlEvents[name] = cls.to_click_control_type(event.__class__)
                for name, sensor in robot.sensors.items():
                    object.sensors[name].types.extend(cls.to_click_control_types(sensor))
            object.objectSensors.append(ValueType.Position)
            object.objectSensors.append(ValueType.RPY)
        return handshake

    @classmethod
    def check_uniform_controltype(cls, objects):
        """
        For backward compatibility we use the specific controlType if all signals are the same type
        """
        robots = list(filter(lambda robot: robot.is_robot(), objects))
        control_types = {type for robot in robots for type in robot.control_types()}
        if len(control_types) == 1:
            return cls.to_click_control_type(robots[0].controlType())
        return ValueType.Multiple

    @classmethod
    def sensor_message_from_objects(cls, robots: List[ClickRobot], simulated_time) -> SensorMessage:
        """
        Create a SensorMessage from a list of robots
        """
        sensor_m = ProtoMessageFactory.create_sensormessage()
        sensor_m.simVars.simulatedTime = simulated_time

        for robot in robots:
            sensors = sensor_m.objects[robot.name]
            if robot.is_robot():
                if len(robot.angle_sensors) > 0:
                    sensors.angleSensors.extend(cls._degs_to_radians(robot.angle_sensors))
                if len(robot.velocity_sensors) > 0:
                    sensors.angleVelocitySensors.extend(cls._signals_to_floats(robot.velocity_sensors))
                if len(robot.torque_sensors) > 0:
                    sensors.torqueSensors.extend(cls._signals_to_floats(robot.torque_sensors))
                cls.add_named_sensors(robot, sensors)
            objectSensor = sensors.objectSensors.add()
            objectSensor.position.arr.extend([*robot.position()])
            objectSensor = sensors.objectSensors.add()
            objectSensor.rpy.arr.extend([*robot.rpy()])

        return sensor_m

    @classmethod
    def add_named_sensors(cls, robot: ClickRobot, sensors):
        for name, signals in robot.sensors.items():
            for signal in signals:
                sensor = sensors.sensors[name].sensor.add()
                value_type = cls.to_click_control_type(signal.__class__)
                if value_type == ValueType.Angle:
                    sensor.angle = signal.GetData()
                elif value_type == ValueType.Torque:
                    sensor.torque = signal.GetData()
                elif value_type == ValueType.AngleVelocity:
                    sensor.angleVelocity = signal.GetData()
                else:
                    arr = cls.array_to_populate(signal, value_type, sensor, robot.name)
                    arr.extend(cls._signal_to_floats(signal))

    @classmethod
    def array_to_populate(cls, signal, value_type, sensor, robotname):
        if value_type == ValueType.Force:
            return sensor.force.arr
        elif value_type == ValueType.DirectionalTorque:
            return sensor.directionalTorque.arr
        elif value_type == ValueType.Angle:
            return sensor.angle
        else:
            # TODO: Add tests for rest of valuetypes, missing 8 so far.
            raise Exception(f"Unrecognized (=Not Implemented) signal in {robotname}: {signal._ModelType}")



from agxclick import ApplicationStepListener, noop
from typing import Callable


class ResetBatchListener(ApplicationStepListener):
    def __init__(self, scene, batch_time: float, on_batch_end: Callable[[], None] = noop) -> None:
        super().__init__()
        self._scene = scene
        self._batch_time = batch_time
        self.batch_start_time = 0.0
        self.current_time = 0.0
        self._on_batch_end = on_batch_end

    def geq(self, a: float, b: float, threshold: float = 0.00001):
        return a > b or (a - b) > -threshold

    def batch_has_ended(self, time):
        return self.geq(time, self.batch_start_time + self._batch_time)

    def prepare_for_next_batch_state(self):
        self.update_parameters()
        self.restart_batch_time(self.current_time)

    def preFrame(self, time: float):
        self.current_time = time
        if self.batch_has_ended(time):
            self.prepare_for_next_batch_state()
            self._on_batch_end()

    def update_parameters(self):
        if self._scene is not None:
            self._scene.UpdateBatchConfiguration()
            self._scene.UpdateParameterSpace()

    def restart_batch_time(self, time: float):
        self.batch_start_time = time
